import io
import os
import platform
import sys

from clikit.api.io.output_stream import OutputStream


class StreamOutputStream(OutputStream):
    """
    An output stream that writes to a stream.
    """

    def __init__(self, stream):  # type: (io.TextIOWrapper) -> None
        self._stream = stream

    def write(self, string):  # type: (str) -> None
        """
        Writes a string to the stream.
        """
        if self.is_closed():
            raise io.UnsupportedOperation("Cannot write to a closed input.")

        self._stream.write(string)

    def flush(self):  # type: () -> None
        """
        Flushes the stream and forces all pending text to be written out.
        """
        if self.is_closed():
            raise io.UnsupportedOperation("Cannot write to a closed input.")

        self._stream.flush()

    def supports_ansi(self):  # type: () -> bool
        """
        Returns whether the stream supports ANSI format codes.
        """
        if platform.system().lower() == "windows":
            shell_supported = (
                os.getenv("ANSICON") is not None
                or "ON" == os.getenv("ConEmuANSI")
                or "xterm" == os.getenv("Term")
            )

            if shell_supported:
                return True

            if not hasattr(self._stream, "fileno"):
                return False

            # Checking for Windows version
            # If we have a compatible version
            # activate color support
            windows_version = sys.getwindowsversion()
            major, build = windows_version[0], windows_version[2]
            if (major, build) < (10, 14393):
                return False

            # Activate colors if possible
            import ctypes
            import ctypes.wintypes

            FILE_TYPE_CHAR = 0x0002
            FILE_TYPE_REMOTE = 0x8000
            ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004

            kernel32 = ctypes.windll.kernel32

            fileno = self._stream.fileno()

            if fileno == 1:
                h = kernel32.GetStdHandle(-11)
            elif fileno == 2:
                h = kernel32.GetStdHandle(-12)
            else:
                return False

            if h is None or h == ctypes.wintypes.HANDLE(-1):
                return False

            if (kernel32.GetFileType(h) & ~FILE_TYPE_REMOTE) != FILE_TYPE_CHAR:
                return False

            mode = ctypes.wintypes.DWORD()
            if not kernel32.GetConsoleMode(h, ctypes.byref(mode)):
                return False

            if (mode.value & ENABLE_VIRTUAL_TERMINAL_PROCESSING) == 0:
                kernel32.SetConsoleMode(
                    h, mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING
                )
                return True

            return False

        if not hasattr(self._stream, "fileno"):
            return False

        try:
            return os.isatty(self._stream.fileno())
        except io.UnsupportedOperation:
            return False

    def close(self):  # type: () -> None
        """
        Closes the stream.
        """
        self._stream.close()

    def is_closed(self):  # type: () -> bool
        """
        Returns whether the stream is closed.
        """
        return self._stream.closed



from .buffered_output_stream import BufferedOutputStream
from .error_output_stream import ErrorOutputStream
from .null_output_stream import NullOutputStream
from .standard_output_stream import StandardOutputStream
from .stream_output_stream import StreamOutputStream



import os

from typing import Optional

from clikit.api.formatter import Formatter
from clikit.formatter import NullFormatter
from clikit.utils._compat import to_str

from .flags import DEBUG
from .flags import NORMAL
from .flags import VERBOSE
from .flags import VERY_VERBOSE
from .output_stream import OutputStream


class Output(Formatter):
    """
    The console output.

    This class wraps an output stream and adds convenience functionality for
    writing that stream.
    """

    def __init__(
        self, stream, formatter=None
    ):  # type: (OutputStream, Optional[Formatter]) -> None
        self._stream = stream

        if formatter is None:
            formatter = NullFormatter()

        self._formatter = formatter
        self._quiet = False
        self._format_output = self._stream.supports_ansi() or formatter.force_ansi()
        self._verbosity = 0

    def write(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a string to the output stream.

        The string is formatted before it is written to the output stream.
        """
        if self._may_write(flags):
            if self._format_output:
                formatted = self.format(string)
            else:
                formatted = self.remove_format(string)

            self._stream.write(to_str(formatted))

    def write_line(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a line of text to the output stream.

        The string is formatted before it is written to the output stream.
        """
        if self._may_write(flags):
            string = string.rstrip(os.linesep)
            if self._format_output:
                formatted = self.format(string)
            else:
                formatted = self.remove_format(string)

            self._stream.write(to_str(formatted + os.linesep))

    def write_raw(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a string to the output stream without formatting.
        """
        if self._may_write(flags):
            self._stream.write(to_str(string))

    def write_line_raw(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a string to the output stream without formatting.
        """
        if self._may_write(flags):
            self._stream.write(to_str(string.rstrip(os.linesep) + os.linesep))

    def flush(self):  # type: () -> None
        """
        Forces all pending text to be written out.
        """
        self._stream.flush()

    def close(self):  # type: () -> None
        """
        Closes the output.
        """
        self._stream.close()

    def is_closed(self):  # type: () -> bool
        """
        Returns whether the output is closed.
        """
        return self._stream.is_closed()

    def set_stream(self, stream):  # type: (OutputStream) -> None
        """
        Sets the underlying stream.
        """
        self._stream = stream
        self._format_output = (
            self._stream.supports_ansi() or self._formatter.force_ansi()
        )

    @property
    def stream(self):  # type: () -> OutputStream
        """
        Returns the underlying stream.
        """
        return self._stream

    def set_formatter(self, formatter):  # type: (Formatter) -> None
        """
        Sets the underlying formatter.
        """
        self._formatter = formatter
        self._format_output = self._stream.supports_ansi() or formatter.force_ansi()

    @property
    def formatter(self):  # type: () -> Formatter
        """
        Returns the underlying formatter.
        """
        return self._formatter

    def set_verbosity(self, verbosity):  # type: (int) -> None
        """
        Sets the verbosity level of the output.
        """
        if verbosity not in {NORMAL, VERBOSE, VERY_VERBOSE, DEBUG}:
            raise ValueError(
                "The verbosity must be one of NORMAL, VERBOSE, VERY_VERBOSE or DEBUG."
            )

        self._verbosity = verbosity

    @property
    def verbosity(self):  # type: () -> int
        """
        Returns the current verbosity level.
        """
        return self._verbosity

    def is_verbose(self):  # type: () -> bool
        """
        Returns whether the verbosity is VERBOSE or greater.
        """
        return self._verbosity >= VERBOSE

    def is_very_verbose(self):  # type: () -> bool
        """
        Returns whether the verbosity is VERY_VERBOSE or greater.
        """
        return self._verbosity >= VERY_VERBOSE

    def is_debug(self):  # type: () -> bool
        """
        Returns whether the verbosity is DEBUG.
        """
        return self._verbosity == DEBUG

    def set_quiet(self, quiet):  # type: (bool) -> None
        """
        Sets whether all output should be suppressed.
        """
        self._quiet = quiet

    def is_quiet(self):  # type: () -> bool
        """
        Returns whether all output is suppressed.
        """
        return self._quiet

    def format(self, string, style=None):  # type: (str, Style) -> str
        """
        Formats the given string.
        """
        return self._formatter.format(string, style)

    def remove_format(self, string):  # type: (str) -> str
        """
        Removes the format tags from the given string.
        """
        return self._formatter.remove_format(string)

    def supports_ansi(self):  # type: () -> bool
        return self._format_output

    def _may_write(self, flags):  # type: (int) -> bool
        """
        Returns whether an output may be written for the given flags.
        """
        if flags is None:
            flags = 0

        if self._quiet:
            return False

        if flags & VERBOSE:
            return self._verbosity >= VERBOSE

        if flags & VERY_VERBOSE:
            return self._verbosity >= VERY_VERBOSE

        if flags & DEBUG:
            return self._verbosity >= DEBUG

        return True



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json

from typing import Any
from typing import Iterable

from clikit.api.args.format import Argument
from clikit.api.args.format import ArgsFormat
from clikit.api.args.format import Option
from clikit.api.io import IO

from clikit.ui import Component
from clikit.ui.components import EmptyLine
from clikit.ui.components import LabeledParagraph
from clikit.ui.components import Paragraph
from clikit.ui.layout import BlockLayout


class AbstractHelp(Component):
    """
    Base class for rendering help pages.
    """

    def render(self, io, indentation=0):  # type: (IO, int) -> None
        layout = BlockLayout()

        self._render_help(layout)

        layout.render(io, indentation)

    def _render_help(self, layout):  # type: (BlockLayout) -> None
        raise NotImplementedError()

    def _render_arguments(
        self, layout, arguments
    ):  # type: (BlockLayout, Iterable[Argument]) -> None
        layout.add(Paragraph("<b>ARGUMENTS</b>"))

        with layout.block():
            for argument in arguments:
                self._render_argument(layout, argument)

        layout.add(EmptyLine())

    def _render_argument(
        self, layout, argument
    ):  # type: (BlockLayout, Argument) -> None
        description = argument.description
        name = "<c1><{}></c1>".format(argument.name)
        default = argument.default

        if default is not None and (not isinstance(default, list) or len(default) > 0):
            description += " <b>{}</b>".format(self._format_value(default))

        layout.add(LabeledParagraph(name, description))

    def _render_options(
        self, layout, options
    ):  # type: (BlockLayout, Iterable[Option]) -> None
        layout.add(Paragraph("<b>OPTIONS</b>"))

        with layout.block():
            for option in options:
                self._render_option(layout, option)

        layout.add(EmptyLine())

    def _render_global_options(
        self, layout, options
    ):  # type: (BlockLayout, Iterable[Option]) -> None
        layout.add(Paragraph("<b>GLOBAL OPTIONS</b>"))

        with layout.block():
            for option in options:
                self._render_option(layout, option)

        layout.add(EmptyLine())

    def _render_option(self, layout, option):  # type: (BlockLayout, Option) -> None
        description = option.description
        default = option.default

        alternative_name = None
        if option.is_long_name_preferred():
            preferred_name = "--{}".format(option.long_name)
            if option.short_name:
                alternative_name = "-{}".format(option.short_name)
        else:
            preferred_name = "-{}".format(option.short_name)
            alternative_name = "--{}".format(option.long_name)

        name = "<c1>{}</c1>".format(preferred_name)
        if alternative_name:
            name += " ({})".format(alternative_name)

        if (
            option.accepts_value()
            and default is not None
            and (not isinstance(default, list) or len(default) > 0)
        ):
            description += " <b>(default: {})</b>".format(self._format_value(default))

        if option.is_multi_valued():
            description = " <b>(multiple values allowed)</b>"

        layout.add(LabeledParagraph(name, description))

    def _render_synopsis(
        self, layout, args_format, app_name, prefix="", last_optional=False
    ):  # type: (BlockLayout, ArgsFormat, str, str, bool) -> None
        name_parts = []
        argument_parts = []

        name_parts.append("<u>{}</u>".format(app_name or "console"))

        for command_name in args_format.get_command_names():
            name_parts.append("<u>{}</u>".format(command_name.string))

        for command_option in args_format.get_command_options():
            if command_option.is_long_name_preferred():
                name_parts.append("--{}".format(command_option.long_name))
            else:
                name_parts.append("-{}".format(command_option.short_name))

        if last_optional:
            name_parts[-1] = "[{}]".format(name_parts[-1])

        for option in args_format.get_options(False).values():
            # \xC2\xA0 is a non-breaking space
            if option.is_value_required():
                fmt = "{}\u00A0<{}>"
            elif option.is_value_optional():
                fmt = "{}\u00A0[<{}>]"
            else:
                fmt = "{}"

            if option.is_long_name_preferred():
                option_name = "--{}".format(option.long_name)
            else:
                option_name = "-{}".format(option.short_name)

            argument_parts.append(
                "[{}]".format(fmt.format(option_name, option.value_name))
            )

        for argument in args_format.get_arguments().values():
            arg_name = argument.name

            argument_parts.append(
                ("<{}>" if argument.is_required() else "[<{}>]").format(
                    arg_name + str(int(argument.is_multi_valued()) or "")
                )
            )

            if argument.is_multi_valued():
                argument_parts.append("... [<{}N>]".format(arg_name))

        args_opts = " ".join(argument_parts)
        name = " ".join(name_parts)

        layout.add(LabeledParagraph(prefix + name, args_opts, 1, False))

    def _format_value(self, value):  # type: (Any) -> str
        return json.dumps(value)



from typing import List

from pylev import levenshtein

from clikit.api.command import CommandCollection


def find_similar_command_names(
    name, commands
):  # type: (str, CommandCollection) -> List[str]
    """
    Finds names similar to a given command name.
    """
    threshold = 1e3
    distance_by_name = {}

    # Include aliases in the search
    actual_names = commands.get_names(True)

    for actual_name in actual_names:
        # Get Levenshtein distance between the input and each command name
        distance = levenshtein(name, actual_name)

        is_similar = distance <= len(name) / 3
        is_sub_string = actual_name.find(name) != -1

        if is_similar or is_sub_string:
            distance_by_name[actual_name] = distance

        # Only keep results with a distance below the threshold
        distance_by_name = {
            k: v for k, v in distance_by_name.items() if v < 2 * threshold
        }

        # Display results with shortest distance first
        suggested_names = []
        for k, v in sorted(distance_by_name.items(), key=lambda _, v: v):
            if k not in suggested_names:
                suggested_names.append(k)

        return suggested_names



from clikit.api.io import OutputStream
from clikit.utils._compat import decode


class BufferedOutputStream(OutputStream):
    """
    An output stream that writes to a buffer.
    """

    def __init__(self):  # type: () -> None
        self._buffer = ""
        self._closed = False

    def fetch(self):  # type: () -> str
        return self._buffer

    def clear(self):  # type: () -> None
        self._buffer = ""

    def write(self, string):  # type: (str) -> None
        """
        Writes a string to the stream.
        """
        if self._closed:
            raise IOError("Cannot read from a closed input.")

        self._buffer += decode(string)

    def flush(self):  # type: () -> None
        """
        Flushes the stream and forces all pending text to be written out.
        """
        if self._closed:
            raise IOError("Cannot read from a closed input.")

    def supports_ansi(self):  # type: () -> bool
        """
        Returns whether the stream supports ANSI format codes.
        """
        return False

    def close(self):  # type: () -> None
        """
        Closes the stream.
        """
        self._closed = True

    def is_closed(self):  # type: () -> bool
        """
        Returns whether the stream is closed.
        """
        return self._closed



from typing import List
from typing import Optional


from clikit.api.args import RawArgs

from .token_parser import TokenParser


class StringArgs(RawArgs):
    """
    Console arguments passed as a string.
    """

    def __init__(self, string):  # type: (str) -> None
        parser = TokenParser()

        self._tokens = parser.parse(string)

    @property
    def script_name(self):  # type: () -> Optional[str]
        return

    @property
    def tokens(self):  # type: () -> List[str]
        return self._tokens

    def has_token(self, token):  # type: (str) -> bool
        return token in self._tokens

    def to_string(self, script_name=True):  # type: (bool) -> str
        return " ".join(self._tokens)



from clikit.api.formatter import Style
from clikit.api.formatter import StyleSet


class DefaultStyleSet(StyleSet):
    """
    The Default TTY style set
    """

    def __init__(self):  # type: () -> None
        styles = [
            Style("info").fg("green"),
            Style("comment").fg("cyan"),
            Style("question").fg("blue"),
            Style("error").fg("red"),
            Style("b").bold(),
            Style("u").underlined(),
            Style("c1").fg("cyan"),
            Style("c2").fg("yellow"),
        ]

        super(DefaultStyleSet, self).__init__(styles)



import re
import time
import threading

from contextlib import contextmanager
from typing import List
from typing import Optional

from clikit.api.io import IO
from clikit.utils.time import format_time


class ProgressIndicator(object):
    """
    A process indicator.
    """

    NORMAL = " {indicator} {message}"
    NORMAL_NO_ANSI = " {message}"
    VERBOSE = " {indicator} {message} ({elapsed:6s})"
    VERBOSE_NO_ANSI = " {message} ({elapsed:6s})"
    VERY_VERBOSE = " {indicator} {message} ({elapsed:6s})"
    VERY_VERBOSE_NO_ANSI = " {message} ({elapsed:6s})"

    def __init__(
        self, io, fmt=None, interval=100, values=None
    ):  # type: (IO, Optional[str], int, Optional[List[str]]) -> None
        self._io = io

        if fmt is None:
            fmt = self._determine_best_format()

        self._fmt = fmt

        if values is None:
            values = ["-", "\\", "|", "/"]

        if len(values) < 2:
            raise ValueError(
                "The progress indicator must have at least 2 indicator value characters."
            )

        self._interval = interval
        self._values = values

        self._message = None
        self._update_time = None
        self._started = False
        self._current = 0

        self._auto_running = None
        self._auto_thread = None

        self._start_time = None
        self._last_message_length = 0

    @property
    def message(self):  # type: () -> Optional[str]
        return self._message

    def set_message(self, message):  # type: (Optional[str]) -> None
        self._message = message

        self._display()

    @property
    def current_value(self):  # type: () -> str
        return self._values[self._current % len(self._values)]

    def start(self, message):  # type: (str) -> None
        if self._started:
            raise RuntimeError("Progress indicator already started.")

        self._message = message
        self._started = True
        self._start_time = time.time()
        self._update_time = self._get_current_time_in_milliseconds() + self._interval
        self._current = 0

        self._display()

    def advance(self):  # type: () -> None
        if not self._started:
            raise RuntimeError("Progress indicator has not yet been started.")

        if not self._io.output.supports_ansi():
            return

        current_time = self._get_current_time_in_milliseconds()
        if current_time < self._update_time:
            return

        self._update_time = current_time + self._interval
        self._current += 1

        self._display()

    def finish(self, message, reset_indicator=False):  # type: (str, bool) -> None
        if not self._started:
            raise RuntimeError("Progress indicator has not yet been started.")

        if self._auto_thread is not None:
            self._auto_running.set()
            self._auto_thread.join()

        self._message = message

        if reset_indicator:
            self._current = 0

        self._display()
        self._io.write_line("")
        self._started = False

    @contextmanager
    def auto(self, start_message, end_message):
        """
        Auto progress.
        """
        self._auto_running = threading.Event()
        self._auto_thread = threading.Thread(target=self._spin)

        self.start(start_message)
        self._auto_thread.start()

        try:
            yield self
        except (Exception, KeyboardInterrupt):
            self._io.write_line("")

            self._auto_running.set()
            self._auto_thread.join()

            raise

        self.finish(end_message, reset_indicator=True)

    def _spin(self):
        while not self._auto_running.is_set():
            self.advance()

            time.sleep(0.1)

    def _display(self):
        if self._io.is_quiet():
            return

        self._overwrite(
            re.sub(
                r"(?i){([a-z\-_]+)(?::([^}]+))?}", self._overwrite_callback, self._fmt
            )
        )

    def _overwrite_callback(self, matches):
        if hasattr(self, "_formatter_{}".format(matches.group(1))):
            text = str(getattr(self, "_formatter_{}".format(matches.group(1)))())
        else:
            text = matches.group(0)

        return text

    def _overwrite(self, message):
        """
        Overwrites a previous message to the output.
        """
        if self._io.output.stream.supports_ansi():
            self._io.write("\x0D\x1B[2K")
            self._io.write(message)
        else:
            self._io.write_line(message)

    def _determine_best_format(self):
        decorated = self._io.output.supports_ansi()

        if self._io.is_very_verbose():
            if decorated:
                return self.VERY_VERBOSE

            return self.VERY_VERBOSE_NO_ANSI
        elif self._io.is_verbose():
            if decorated:
                return self.VERY_VERBOSE

            return self.VERBOSE_NO_ANSI

        if decorated:
            return self.NORMAL

        return self.NORMAL_NO_ANSI

    def _get_current_time_in_milliseconds(self):
        return round(time.time() * 1000)

    def _formatter_indicator(self):
        return self.current_value

    def _formatter_message(self):
        return self.message

    def _formatter_elapsed(self):
        return format_time(time.time() - self._start_time)



from .null_input_stream import NullInputStream
from .standard_input_stream import StandardInputStream
from .stream_input_stream import StreamInputStream
from .string_input_stream import StringInputStream



import re

from typing import Any
from typing import Optional

from clikit.utils._compat import basestring
from clikit.utils.string import parse_boolean
from clikit.utils.string import parse_float
from clikit.utils.string import parse_int
from clikit.utils.string import parse_string


class Argument(object):
    """
    An input argument.
    """

    REQUIRED = 1
    OPTIONAL = 2
    MULTI_VALUED = 4
    STRING = 16
    BOOLEAN = 32
    INTEGER = 64
    FLOAT = 128
    NULLABLE = 256

    def __init__(
        self, name, flags=0, description=None, default=None
    ):  # type: (str, int, Optional[str], Any) -> None
        if not isinstance(name, basestring):
            raise ValueError(
                "The argument name must be a string. Got: {}".format(type(name))
            )

        if not name:
            raise ValueError("The argument name must not be empty.")

        if not name[:1].isalpha():
            raise ValueError("The argument name must start with a letter")

        if not re.match(r"^[a-zA-Z0-9\-]+$", name):
            raise ValueError(
                "The argument name must contain letters, digits and hyphens only."
            )

        if description is not None:
            if not isinstance(description, basestring):
                raise ValueError(
                    "The argument description must be a string. Got: {}".format(
                        type(description)
                    )
                )

            if not description:
                raise ValueError("The argument description must not be empty.")

        self._validate_flags(flags)

        flags = self._add_default_flags(flags)

        self._name = name
        self._flags = flags
        self._description = description

        if self.is_multi_valued():
            self._default = []
        else:
            self._default = None

        if self.is_optional() or default is not None:
            self.set_default(default)

    @property
    def name(self):  # type: () -> str
        return self._name

    @property
    def flags(self):  # type: () -> int
        return self._flags

    @property
    def description(self):  # type: () -> Optional[str]
        return self._description

    @property
    def default(self):  # type: () -> Any
        return self._default

    def is_required(self):  # type: () -> bool
        return bool(self.REQUIRED & self._flags)

    def is_optional(self):  # type: () -> bool
        return bool(self.OPTIONAL & self._flags)

    def is_multi_valued(self):  # type: () -> bool
        return bool(self.MULTI_VALUED & self._flags)

    def set_default(self, default=None):  # type: (Any) -> None
        if self.is_required():
            raise ValueError("Required arguments do not accept default values.")

        if self.is_multi_valued():
            if default is None:
                default = []
            elif not isinstance(default, list):
                raise ValueError(
                    "The default value of a multi-valued argument must be a list. "
                    "Got: {}".format(type(default))
                )

        self._default = default

    def parse(self, value):  # type: (Any) -> Any
        nullable = bool(self._flags & self.NULLABLE)

        if self._flags & self.BOOLEAN:
            return parse_boolean(value, nullable)
        elif self._flags & self.INTEGER:
            return parse_int(value, nullable)
        elif self._flags & self.FLOAT:
            return parse_float(value, nullable)

        return parse_string(value, nullable)

    def _validate_flags(self, flags):  # type: (int) -> None
        if flags & self.REQUIRED and flags & self.OPTIONAL:
            raise ValueError(
                "The argument flags REQUIRED and OPTIONAL cannot be combined."
            )

        if flags & self.STRING:
            if flags & self.BOOLEAN:
                raise ValueError(
                    "The argument flags STRING and BOOLEAN cannot be combined."
                )

            if flags & self.INTEGER:
                raise ValueError(
                    "The argument flags STRING and INTEGER cannot be combined."
                )

            if flags & self.FLOAT:
                raise ValueError(
                    "The argument flags STRING and FLOAT cannot be combined."
                )
        elif flags & self.BOOLEAN:
            if flags & self.INTEGER:
                raise ValueError(
                    "The argument flags BOOLEAN and INTEGER cannot be combined."
                )

            if flags & self.FLOAT:
                raise ValueError(
                    "The argument flags BOOLEAN and FLOAT cannot be combined."
                )
        elif flags & self.INTEGER:
            if flags & self.FLOAT:
                raise ValueError(
                    "The argument flags INTEGER and FLOAT cannot be combined."
                )

    def _add_default_flags(self, flags):  # type: (int) -> int
        if not flags & (self.REQUIRED | self.OPTIONAL):
            flags |= self.OPTIONAL

        if not flags & (self.STRING | self.BOOLEAN | self.INTEGER | self.FLOAT):
            flags |= self.STRING

        return flags



from .api.config.application_config import ApplicationConfig
from .console_application import ConsoleApplication
from .config.default_application_config import DefaultApplicationConfig


__version__ = "0.1.0"



from clikit.api.application import Application
from clikit.api.args.format import ArgsFormat
from clikit.api.args.format import ArgsFormatBuilder
from clikit.api.args.format import Argument
from clikit.api.command import Command
from clikit.api.command import CommandCollection
from clikit.ui.components import EmptyLine
from clikit.ui.components import LabeledParagraph
from clikit.ui.components import NameVersion
from clikit.ui.components import Paragraph
from clikit.ui.layout import BlockLayout

from .abstract_help import AbstractHelp


class ApplicationHelp(AbstractHelp):
    """
    Renders the help of a console application.
    """

    def __init__(self, application):  # type: (Application) -> None
        self._application = application

    def _render_help(self, layout):  # type: (BlockLayout) -> None
        help = self._application.config.help
        commands = self._application.named_commands
        global_args_format = self._application.global_args_format

        builder = ArgsFormatBuilder()
        builder.add_argument(
            Argument("command", Argument.REQUIRED, "The command to execute")
        )
        builder.add_argument(
            Argument("arg", Argument.MULTI_VALUED, "The arguments of the command")
        )
        builder.add_options(*global_args_format.get_options().values())

        args_format = builder.format

        self._render_name(layout, self._application)
        self._render_usage(layout, self._application, args_format)
        self._render_arguments(layout, args_format.get_arguments().values())

        if args_format.has_options():
            self._render_global_options(layout, args_format.get_options().values())

        if not commands.is_empty():
            self._render_commands(layout, commands)

        if help:
            self._render_description(layout, help)

    def _render_name(
        self, layout, application
    ):  # type: (BlockLayout, Application) -> None
        layout.add(NameVersion(application.config))
        layout.add(EmptyLine())

    def _render_usage(
        self, layout, application, args_format
    ):  # type: (BlockLayout, Application, ArgsFormat) -> None
        app_name = application.config.name

        layout.add(Paragraph("<b>USAGE</b>"))

        with layout.block():
            self._render_synopsis(layout, args_format, app_name)

        layout.add(EmptyLine())

    def _render_commands(
        self, layout, commands
    ):  # type: (BlockLayout, CommandCollection) -> None
        layout.add(Paragraph("<b>AVAILABLE COMMANDS</b>"))

        with layout.block():
            for command in sorted(commands, key=lambda c: c.name):
                self._render_command(layout, command)

        layout.add(EmptyLine())

    def _render_command(self, layout, command):  # type: (BlockLayout, Command) -> None
        description = command.config.description
        name = "<c1>{}</c1>".format(command.name)

        layout.add(LabeledParagraph(name, description))

    def _render_description(self, layout, help):  # type: (BlockLayout, str) -> None
        layout.add(Paragraph("<b>DESCRIPTION</b>"))
        with layout.block():
            layout.add(Paragraph(help))

        layout.add(EmptyLine())



from typing import Optional


class Style(object):
    """
    A formatter style.
    """

    _colors = {"black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"}

    def __init__(self, tag=None):  # type: (Optional[str]) -> None
        self._tag = tag
        self._fg_color = None
        self._bg_color = None
        self._bold = False
        self._underlined = False
        self._blinking = False
        self._inverse = False
        self._hidden = False

    @property
    def tag(self):  # type: () -> Optional[str]
        return self._tag

    @property
    def foreground_color(self):  # type: () -> str
        return self._fg_color

    @property
    def background_color(self):  # type: () -> str
        return self._bg_color

    def fg(self, color):  # type: (str) -> Style
        """
        Sets the foreground color.
        """
        self._fg_color = color

        return self

    def bg(self, color):  # type: (str) -> Style
        """
        Sets the background color.
        """
        self._bg_color = color

        return self

    def bold(self, bold=True):  # type: (bool) -> Style
        """
        Sets or unsets the font weight to bold.
        """
        self._bold = bold

        return self

    def underlined(self, underlined=True):  # type: (bool) -> Style
        """
        Enables or disables underlining.
        """
        self._underlined = underlined

        return self

    def blinking(self, blinking=True):  # type: (bool) -> Style
        """
        Enables or disables blinking.
        """
        self._blinking = blinking

        return self

    def inverse(self, inverse=True):  # type: (bool) -> Style
        """
        Enables or disables inverse colors.
        """
        self._inverse = inverse

        return self

    def hidden(self, hidden=True):  # type: (bool) -> Style
        """
        Hides or shows the text.
        """
        self._hidden = hidden

        return self

    def is_bold(self):  # type: () -> bool
        return self._bold

    def is_underlined(self):  # type: () -> bool
        return self._underlined

    def is_blinking(self):  # type: () -> bool
        return self._blinking

    def is_inverse(self):  # type: () -> bool
        return self._inverse

    def is_hidden(self):  # type: () -> bool
        return self._hidden



from .args_format import ArgsFormat
from .args_format_builder import ArgsFormatBuilder
from .argument import Argument
from .command_name import CommandName
from .command_option import CommandOption
from .option import Option



from clikit.api.args import Args
from clikit.api.command import Command
from clikit.api.io import IO

from clikit.ui.help import ApplicationHelp
from clikit.ui.help import CommandHelp


class HelpTextHandler:
    """
    Displays help as text format.
    """

    def handle(self, args, io, command):  # type: (Args, IO, Command) -> int
        application = command.application

        if args.is_argument_set("command"):
            the_command = application.get_command(args.argument("command"))

            usage = CommandHelp(the_command)
        else:
            usage = ApplicationHelp(application)

        usage.render(io)

        return 0



from typing import Optional


from clikit.api.formatter import Formatter
from clikit.api.io import IO
from clikit.api.io import Input
from clikit.api.io import Output
from clikit.formatter import PlainFormatter
from clikit.utils._compat import to_str

from .input_stream import StringInputStream
from .output_stream import BufferedOutputStream


class BufferedIO(IO):
    """
    An I/O that reads from and writes to a buffer.
    """

    def __init__(
        self, input_data="", formatter=None
    ):  # type: (str, Optional[Formatter]) -> None
        if formatter is None:
            formatter = PlainFormatter()

        input = Input(StringInputStream(input_data))
        output = Output(BufferedOutputStream(), formatter)
        error_output = Output(BufferedOutputStream(), formatter)

        super(BufferedIO, self).__init__(input, output, error_output)

    def set_input(self, data):  # type: (str) -> None
        self.input.stream.set(data)

    def append_input(self, data):  # type: (str) -> None
        self.input.stream.append(data)

    def clear_input(self):  # type: () -> None
        self.input.stream.clear()

    def fetch_output(self):  # type: () -> str
        return self.output.stream.fetch()

    def clear_output(self):  # type: () -> None
        self.output.stream.clear()

    def fetch_error(self):  # type: () -> str
        return self.error_output.stream.fetch()

    def clear_error(self):  # type: () -> None
        self.error_output.stream.clear()



# -*- coding: utf-8 -*-

import os
import platform
import struct
import subprocess
import shlex


class Terminal(object):
    """
    Represents the current terminal.
    """

    def __init__(self):
        self._width = None
        self._height = None

    @property
    def width(self):
        width = os.getenv("COLUMNS", "").strip()
        if width:
            return int(width)

        if self._width is None:
            self._init_dimensions()

        return self._width

    @property
    def height(self):
        height = os.getenv("LINES", "").strip()
        if height:
            return int(height)

        if self._height is None:
            self._init_dimensions()

        return self._height

    def _init_dimensions(self):
        current_os = platform.system().lower()
        dimensions = None

        if current_os.lower() == "windows":
            dimensions = self._get_terminal_size_windows()
            if dimensions is None:
                dimensions = self._get_terminal_size_tput()
        elif current_os.lower() in ["linux", "darwin"] or current_os.startswith(
            "cygwin"
        ):
            dimensions = self._get_terminal_size_linux()

        if dimensions is None:
            dimensions = 80, 25

        self._width, self._height = dimensions

    def _get_terminal_size_windows(self):
        try:
            from ctypes import windll, create_string_buffer

            # stdin handle is -10
            # stdout handle is -11
            # stderr handle is -12
            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
            if res:
                (
                    bufx,
                    bufy,
                    curx,
                    cury,
                    wattr,
                    left,
                    top,
                    right,
                    bottom,
                    maxx,
                    maxy,
                ) = struct.unpack("hhhhHhhhhhh", csbi.raw)
                sizex = right - left + 1
                sizey = bottom - top + 1
                return sizex, sizey
        except:
            pass

    def _get_terminal_size_tput(self):
        # get terminal width
        # src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window
        try:
            cols = int(
                subprocess.check_output(
                    shlex.split("tput cols"), stderr=subprocess.STDOUT
                )
            )
            rows = int(
                subprocess.check_output(
                    shlex.split("tput lines"), stderr=subprocess.STDOUT
                )
            )

            return (cols, rows)
        except:
            pass

    def _get_terminal_size_linux(self):
        def ioctl_GWINSZ(fd):
            try:
                import fcntl
                import termios

                cr = struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
                return cr
            except:
                pass

        cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
        if not cr:
            try:
                fd = os.open(os.ctermid(), os.O_RDONLY)
                cr = ioctl_GWINSZ(fd)
                os.close(fd)
            except:
                pass

        if not cr:
            try:
                cr = (os.environ["LINES"], os.environ["COLUMNS"])
            except:
                return None

        return int(cr[1]), int(cr[0])



import re

from typing import Optional


from clikit.utils._compat import basestring


class AbstractOption(object):
    """
    Base class for command line options.
    """

    PREFER_LONG_NAME = 1
    PREFER_SHORT_NAME = 2

    def __init__(
        self, long_name, short_name=None, flags=0, description=None
    ):  # type: (str, Optional[str], int, Optional[str]) -> None
        long_name = self._remove_double_dash_prefix(long_name)
        short_name = self._remove_dash_prefix(short_name)

        self._validate_flags(flags)
        self._validate_long_name(long_name)
        self._validate_short_name(short_name, flags)

        self._long_name = long_name
        self._short_name = short_name
        self._description = description

        flags = self._add_default_flags(flags)

        self._flags = flags

    @property
    def long_name(self):  # type: () -> str
        return self._long_name

    @property
    def short_name(self):  # type: () -> Optional[str]
        return self._short_name

    @property
    def flags(self):  # type: () -> int
        return self._flags

    @property
    def description(self):  # type: () -> Optional[str]
        return self._description

    def is_long_name_preferred(self):  # type: () -> bool
        return bool(self._flags & self.PREFER_LONG_NAME)

    def is_short_name_preferred(self):  # type: () -> bool
        return bool(self._flags & self.PREFER_SHORT_NAME)

    def _remove_double_dash_prefix(self, string):  # type: (str) -> str
        if not isinstance(string, basestring):
            return string

        if string.startswith("--"):
            string = string[2:]

        return string

    def _remove_dash_prefix(self, string):  # type: (Optional[str]) -> Optional[str]
        if string is None:
            return string

        if not isinstance(string, basestring):
            return string

        if string.startswith("-"):
            string = string[1:]

        return string

    def _validate_flags(self, flags):  # type: (int) -> None
        if flags & self.PREFER_SHORT_NAME and flags & self.PREFER_LONG_NAME:
            raise ValueError(
                "The option flags PREFER_SHORT_NAME and PREFER_LONG_NAME cannot be combined."
            )

    def _validate_long_name(self, long_name):  # type: (Optional[str]) -> None
        if long_name is None:
            raise ValueError("The long option name must not be null.")

        if not isinstance(long_name, basestring):
            raise ValueError(
                "The long option name must be a string. Got: {}".format(type(long_name))
            )

        if not long_name:
            raise ValueError("The long option name must not be empty.")

        if len(long_name) < 2:
            raise ValueError(
                'The long option name must contain more than one character. Got: "{}"'.format(
                    len(long_name)
                )
            )

        if not long_name[:1].isalpha():
            raise ValueError("The long option name must start with a letter")

        if not re.match(r"^[a-zA-Z0-9\-]+$", long_name):
            raise ValueError(
                "The long option name must contain letters, digits and hyphens only."
            )

    def _validate_short_name(
        self, short_name, flags
    ):  # type: (Optional[str], int) -> None
        if short_name is None:
            if flags & self.PREFER_SHORT_NAME:
                raise ValueError(
                    "The short option name must be given if the option flag PREFER_SHORT_NAME is selected."
                )

            return

        if not isinstance(short_name, basestring):
            raise ValueError(
                "The short option name must be a string. Got: {}".format(
                    type(short_name)
                )
            )

        if not short_name:
            raise ValueError("The short option name must not be empty.")

        if not re.match(r"^[a-zA-Z]$", short_name):
            raise ValueError("The short option name must be exactly one letter.")

    def _add_default_flags(self, flags):  # type: (int) -> int
        if not flags & (self.PREFER_LONG_NAME | self.PREFER_SHORT_NAME):
            flags |= (
                self.PREFER_SHORT_NAME if self._short_name else self.PREFER_LONG_NAME
            )

        return flags



import sys

from .stream_output_stream import StreamOutputStream


class ErrorOutputStream(StreamOutputStream):
    """
    An output stream that writes to the error output.
    """

    def __init__(self):  # type: () -> None
        super(ErrorOutputStream, self).__init__(sys.stderr)



import sys

from .stream_input_stream import StreamInputStream


class StandardInputStream(StreamInputStream):
    """
    An input stream that reads from the standard input.
    """

    def __init__(self):  # type: () -> None
        super(StandardInputStream, self).__init__(sys.stdin)



from typing import Iterable

from clikit.api.args.format import Argument
from clikit.api.args.format import Option
from clikit.api.command import Command
from clikit.api.command import CommandCollection
from clikit.ui.components import EmptyLine
from clikit.ui.components import Paragraph
from clikit.ui.layout import BlockLayout

from .abstract_help import AbstractHelp


class CommandHelp(AbstractHelp):
    """
    Renders the help of a command.
    """

    def __init__(self, command):  # type: (Command) -> None
        self._command = command

    def _render_help(self, layout):  # type: (BlockLayout) -> None
        help = self._command.config.help
        args_format = self._command.args_format
        sub_commands = self._command.named_sub_commands

        self._render_usage(layout, self._command)

        if args_format.has_arguments():
            self._render_arguments(layout, args_format.get_arguments().values())

        if not sub_commands.is_empty():
            self._render_sub_commands(layout, sub_commands)

        if args_format.has_options(False):
            self._render_options(layout, args_format.get_options(False).values())

        if args_format.base_format and args_format.base_format.has_options():
            self._render_global_options(
                layout, args_format.base_format.get_options().values()
            )

        if help:
            self._render_description(layout, help)

    def _render_usage(self, layout, command):  # type: (BlockLayout, Command) -> None
        formats_to_print = []

        # Start with the default commands
        if command.has_default_sub_commands():
            # If the command has default commands, print them
            for sub_command in command.default_sub_commands:
                # The name of the sub command is only optional (i.e. printed
                # wrapped in brackets: "[sub]") if the command is not
                # anonymous
                name_optional = not sub_command.config.is_anonymous()

                formats_to_print.append((sub_command.args_format, name_optional))
        else:
            # Otherwise print the command's usage itself
            formats_to_print.append((command.args_format, False))

        # Add remaining sub-commands
        for sub_command in command.sub_commands:
            # Don't duplicate default commands
            if not sub_command.config.is_default():
                formats_to_print.append((sub_command.args_format, False))

        app_name = command.application.config.name
        prefix = "    " if len(formats_to_print) > 1 else ""

        layout.add(Paragraph("<b>USAGE</b>"))
        with layout.block():
            for vars in formats_to_print:
                self._render_synopsis(layout, vars[0], app_name, prefix, vars[1])
                prefix = "or: "

            if command.has_aliases():
                layout.add(EmptyLine())
                self._render_aliases(layout, command.aliases)

        layout.add(EmptyLine())

    def _render_aliases(
        self, layout, aliases
    ):  # type: (BlockLayout, Iterable[str]) -> None
        layout.add(Paragraph("aliases: {}".format(", ".join(aliases))))

    def _render_sub_commands(
        self, layout, sub_commands
    ):  # type: (BlockLayout, CommandCollection) -> None
        layout.add(Paragraph("<b>COMMANDS</b>"))

        with layout.block():
            for sub_command in sorted(sub_commands, key=lambda c: c.name):
                self._render_sub_command(layout, sub_command)

    def _render_sub_command(
        self, layout, command
    ):  # type: (BlockLayout, Command) -> None
        config = command.config
        description = config.description
        help = config.help
        arguments = command.args_format.get_arguments(False)
        options = command.args_format.get_options(False)

        # TODO: option commands
        name = "<u>{}</u>".format(command.name)

        layout.add(Paragraph(name))

        with layout.block():
            if description:
                self._render_sub_command_description(layout, description)

            if help:
                self._render_sub_command_help(layout, help)

            if arguments:
                self._render_sub_command_arguments(layout, arguments.values())

            if options:
                self._render_sub_command_options(layout, options.values())

            if not description and not help and not arguments and not options:
                layout.add(EmptyLine())

    def _render_sub_command_description(
        self, layout, description
    ):  # type: (BlockLayout, str) -> None
        layout.add(Paragraph(description))
        layout.add(EmptyLine())

    def _render_sub_command_help(
        self, layout, help
    ):  # type: (BlockLayout, str) -> None
        layout.add(Paragraph(help))
        layout.add(EmptyLine())

    def _render_sub_command_arguments(
        self, layout, arguments
    ):  # type: (BlockLayout, Iterable[Argument]) -> None
        for argument in arguments:
            self._render_argument(layout, argument)

        layout.add(EmptyLine())

    def _render_sub_command_options(
        self, layout, options
    ):  # type: (BlockLayout, Iterable[Option]) -> None
        for option in options:
            self._render_option(layout, option)

        layout.add(EmptyLine())

    def _render_description(self, layout, help):  # type: (BlockLayout, str) -> None
        layout.add(Paragraph("<b>DESCRIPTION</b>"))
        with layout.block():
            layout.add(Paragraph(help))

        layout.add(EmptyLine())



from typing import Optional

from clikit.api.formatter import Formatter

from .input import Input
from .output import Output


class IO(Formatter):
    """
    Provides methods to access the console input and output.
    """

    def __init__(
        self, input, output, error_output
    ):  # type: (Input, Output, Output) -> None
        self._input = input
        self._output = output
        self._error_output = error_output
        self._terminal_dimensions = None

    @property
    def input(self):  # type: () -> Input
        return self._input

    @property
    def output(self):  # type: () -> Output
        return self._output

    @property
    def error_output(self):  # type: () -> Output
        return self._error_output

    def read(self, length, default=None):  # type: (int, Optional[str]) -> str
        """
        Reads the given amount of characters from the standard input.

        :raises: IOException
        """
        return self._input.read(length, default=default)

    def read_line(
        self, length=None, default=None
    ):  # type: (Optional[int], Optional[str]) -> str
        """
        Reads a line from the standard input.

        :raises: IOException
        """
        return self._input.read_line(length=length, default=default)

    def write(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a string to the standard output.

        The string is formatted before it is written to the output.
        """
        self._output.write(string, flags=flags)

    def write_line(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a line to the standard output.

        The string is formatted before it is written to the output.
        """
        self._output.write_line(string, flags=flags)

    def write_raw(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a string to the standard output without formatting.
        """
        self._output.write_raw(string, flags=flags)

    def write_line_raw(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a line to the standard output without formatting.
        """
        self._output.write_raw(string, flags=flags)

    def error(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a string to the error output.

        The string is formatted before it is written to the output.
        """
        self._error_output.write(string, flags=flags)

    def error_line(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a line to the error output.

        The string is formatted before it is written to the output.
        """
        self._error_output.write_line(string, flags=flags)

    def error_raw(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a string to the error output without formatting.
        """
        self._error_output.write_raw(string, flags=flags)

    def error_line_raw(self, string, flags=None):  # type: (str, Optional[int]) -> None
        """
        Writes a line to the error output without formatting.
        """
        self._error_output.write_raw(string, flags=flags)

    def flush(self):  # type: () -> None
        """
        Flushes the outputs and forces all pending text to be written out.
        """
        self._output.flush()
        self._error_output.flush()

    def close(self):  # type: () -> None
        """
        Closes the input and the outputs.
        """
        self._input.close()
        self._output.close()
        self._error_output.close()

    def set_interactive(self, interactive):  # type: (bool) -> None
        """
        Enables or disables interaction with the user.
        """
        self._input.set_interactive(interactive)

    def is_interactive(self):  # type: () -> bool
        """
        Returns whether the user may be asked for input.
        """
        return self._input.is_interactive()

    def set_verbosity(self, verbosity):  # type: (int) -> None
        """
        Sets the verbosity of the output.
        """
        self._output.set_verbosity(verbosity)
        self._error_output.set_verbosity(verbosity)

    def is_verbose(self):  # type: () -> bool
        """
        Returns whether the verbosity is VERBOSE or greater.
        """
        return self._output.is_verbose()

    def is_very_verbose(self):  # type: () -> bool
        """
        Returns whether the verbosity is VERY_VERBOSE or greater.
        """
        return self._output.is_very_verbose()

    def is_debug(self):  # type: () -> bool
        """
        Returns whether the verbosity is DEBUG.
        """
        return self._output.is_debug()

    @property
    def verbosity(self):  # type: () -> int
        return self._output.verbosity

    def set_quiet(self, quiet):  # type: (bool) -> None
        """
        Sets whether all output should be suppressed.
        """
        self._output.set_quiet(quiet)
        self._error_output.set_quiet(quiet)

    def is_quiet(self):  # type: () -> bool
        """
        Returns whether all output is suppressed.
        """
        return self._output.is_quiet()

    def set_terminal_dimensions(self, dimensions):  # type: (Rectangle) -> None
        """
        Sets the dimensions of the terminal.
        """
        self._terminal_dimensions = dimensions

    @property
    def terminal_dimensions(self):  # type: () -> Rectangle
        if not self._terminal_dimensions:
            self._terminal_dimensions = self.get_default_terminal_dimensions()

        return self._terminal_dimensions

    def get_default_terminal_dimensions(self):  # type: () -> Rectangle
        """
        Returns the default terminal dimensions.
        """
        from clikit.ui.rectangle import Rectangle

        return Rectangle(80, 20)

    def set_formatter(self, formatter):  # type: (Formatter) -> None
        """
        Sets the output formatter.
        """
        self._output.set_formatter(formatter)
        self._error_output.set_formatter(formatter)

    @property
    def formatter(self):  # type: () -> Formatter
        """
        Returns the output formatter.
        """
        return self._output.formatter

    def format(self, string, style=None):  # type: (str, Style) -> str
        """
        Formats the given string.
        """
        return self._output.formatter.format(string, style=style)

    def remove_format(self, string):  # type: (str) -> str
        """
        Removes the format tags from the given string.
        """
        return self._output.formatter.remove_format(string)



from typing import Optional

from clikit.api.formatter import Formatter


class NullFormatter(Formatter):
    """
    A formatter that returns all text unchanged.
    """

    def format(self, string, style=None):  # type: (str, Optional[Style]) -> str
        return string

    def remove_format(self, string):  # type: (str) -> str
        return string

    def force_ansi(self):  # type: () -> bool
        return False



from __future__ import unicode_literals

import math

from typing import List

from clikit.api.formatter import Style
from clikit.api.io import IO
from clikit.ui.style.alignment import Alignment
from clikit.ui.style.border_style import BorderStyle
from clikit.utils.string import get_string_length


class BorderUtil:
    """
    Contains utility methods to draw borders and bordered cells.
    """

    @classmethod
    def draw_top_border(
        cls, io, style, column_lengths, indentation=0
    ):  # type: (IO, BorderStyle, List[int], int) -> None
        cls.draw_border(
            io,
            column_lengths,
            indentation,
            style.line_ht_char,
            style.corner_tl_char,
            style.crossing_t_char,
            style.corner_tr_char,
            style.style,
        )

    @classmethod
    def draw_middle_border(
        cls, io, style, column_lengths, indentation=0
    ):  # type: (IO, BorderStyle, List[int], int) -> None
        cls.draw_border(
            io,
            column_lengths,
            indentation,
            style.line_hc_char,
            style.crossing_l_char,
            style.crossing_c_char,
            style.crossing_r_char,
            style.style,
        )

    @classmethod
    def draw_bottom_border(
        cls, io, style, column_lengths, indentation=0
    ):  # type: (IO, BorderStyle, List[int], int) -> None
        cls.draw_border(
            io,
            column_lengths,
            indentation,
            style.line_hb_char,
            style.corner_bl_char,
            style.crossing_b_char,
            style.corner_br_char,
            style.style,
        )

    @classmethod
    def draw_row(
        cls,
        io,
        style,
        row,
        column_lengths,
        alignments,
        cell_format,
        padding_char,
        cell_style=None,
        indentation=0,
    ):  # type: (IO, BorderStyle, List[str], List[int], List[int], str, str, Style, int) -> None
        total_lines = 0

        # Split all cells into lines
        for col, cell in enumerate(row):
            row[col] = cell.split("\n")
            total_lines = max(total_lines, len(row[col]))

        nb_columns = len(row)
        border_vl_char = io.format(style.line_vl_char, style.style)
        border_vc_char = io.format(style.line_vc_char, style.style)
        border_vr_char = io.format(style.line_vr_char, style.style)

        for i in range(total_lines):
            line = " " * indentation
            line += border_vl_char

            for col, remaining_lines in enumerate(row):
                if remaining_lines:
                    cell_line = remaining_lines.pop(0)
                else:
                    cell_line = ""

                total_pad_length = column_lengths[col] - get_string_length(
                    cell_line, io
                )
                padding_left = ""
                padding_right = ""

                if total_pad_length >= 0:
                    try:
                        alignment = alignments[col]
                    except IndexError:
                        alignment = Alignment.LEFT

                    if alignment == Alignment.LEFT:
                        padding_right = padding_char * total_pad_length
                    elif alignment == Alignment.RIGHT:
                        padding_left = padding_char * total_pad_length
                    else:
                        left_pad_length = int(math.floor(total_pad_length / 2))
                        padding_left = padding_char * left_pad_length
                        padding_right = padding_char * (
                            total_pad_length - left_pad_length
                        )

                    line += io.format(
                        cell_format.format(padding_left + cell_line + padding_right),
                        cell_style,
                    )
                    if col < nb_columns - 1:
                        line += border_vc_char
                    else:
                        line += border_vr_char

            # Remove trailing space
            io.write(line.rstrip() + "\n")

    @classmethod
    def draw_border(
        cls,
        io,
        column_lengths,
        indentation,
        line_char,
        crossing_l_char,
        crossing_c_char,
        crossing_r_char,
        style=None,
    ):  # type: (IO, List[int], int, str, str, str, str, Style) -> None
        line = " " * indentation
        line += crossing_l_char

        l = len(column_lengths)
        for i in range(l):
            line += line_char * column_lengths[i]
            if i < l - 1:
                line += crossing_c_char
            else:
                line += crossing_r_char

        line = line.rstrip()

        if line:
            io.write(io.format(line, style) + "\n")
